/// gemtext dart implementation
library gemtext;

/// MIME type of gemtext file
const gemtextMime = "text/gemini";

/// gemtext file extensions
const gemtextFileExtensions = [".gmi", ".gemini"];

/// A gemtext document
class Gemtext {
  final List<Node> nodes;

  Gemtext([List<Node>? nodes]) : nodes = nodes ?? [];

  factory Gemtext.parse(String text) {
    bool preformatted = false;
    StringBuffer preformattedBuffer = StringBuffer();
    String? alt;

    final List<Node> res = [];

    for (final line in text.split("\n")) {
      if (line.startsWith("```")) {
        if (!preformatted) {
          alt = line.trim() == "```" ? null : line.trim().substring(3).trim();
        } else {
          final str = preformattedBuffer.toString();
          res.add(Preformatted(alt, str.substring(0, str.length - 1)));
          alt = null;
          preformattedBuffer.clear();
        }
        preformatted = !preformatted;
      } else if (preformatted) {
        preformattedBuffer.writeln(line);
      } else {
        if (line.startsWith("=>")) {
          final stripped = line.substring(2).trim().split(RegExp(r"\s"));
          final to = Uri.parse(stripped[0].trim());
          final name = stripped.length == 1
              ? null
              : stripped.sublist(1).join(' ').trim();
          res.add(Link(to, name));
        } else if (line.startsWith(">")) {
          res.add(Quote(line.substring(1).trim()));
        } else if (line.startsWith("# ")) {
          res.add(Heading(HeadingLevel.h1, line.substring(2).trim()));
        } else if (line.startsWith("## ")) {
          res.add(Heading(HeadingLevel.h2, line.substring(3).trim()));
        } else if (line.startsWith("### ")) {
          res.add(Heading(HeadingLevel.h3, line.substring(4).trim()));
        } else if (line.startsWith("* ")) {
          res.add(ListItem(line.substring(2).trim()));
        } else {
          res.add(Text(line.trim()));
        }
      }
    }

    return Gemtext(res);
  }

  void push(Node node) => nodes.add(node);
  void text(String text) => push(Text(text));
  void heading(HeadingLevel level, String text) => push(Heading(level, text));
  void h1(String text) => heading(HeadingLevel.h1, text);
  void h2(String text) => heading(HeadingLevel.h2, text);
  void h3(String text) => heading(HeadingLevel.h3, text);
  void link(Uri to, [String? name]) => push(Link(to, name));
  void quote(String body) => push(Quote(body));
  void listItem(String body) => push(ListItem(body));
  void preformatted(String text, [String? alt]) =>
      push(Preformatted(alt, text));

  @override
  String toString() => nodes.join("\n");
}

/// Individual nodes of the document. Each node correlates to a line in the file.
sealed class Node {}

/// Lines beginning with "#" are heading lines. Heading lines consist of one,
/// two or three consecutive "#" characters, followed by optional whitespace,
/// followed by heading text. The number of # characters indicates the "level"
/// of header; #, ## and ### can be thought of as analogous to `<h1>`, `<h2>`
/// and `<h3>` in HTML.
///
/// Heading text should be presented to the user, and clients MAY use special
/// formatting, e.g. a larger or bold font, to indicate its status as a header
/// (simple clients may simply print the line, including its leading #s,
/// without any styling at all). However, the main motivation for the
/// definition of heading lines is not stylistic but to provide a
/// machine-readable representation of the internal structure of the document.
/// Advanced clients can use this information to, e.g. display an automatically
/// generated and hierarchically formatted "table of contents" for a long
/// document in a side-pane, allowing users to easily jump to specific sections
/// without excessive scrolling. CMS-style tools automatically generating menus
/// or Atom/RSS feeds for a directory of text/gemini files can use first
/// heading in the file as a human-friendly title.
class Heading extends Node {
  final HeadingLevel level;
  final String body;

  Heading(this.level, this.body);

  @override
  String toString() => "${'#' * (level.index + 1)} $body";
}

enum HeadingLevel { h1, h2, h3 }

/// Lines beginning with the two characters "=>" are link lines, which have the following syntax:
///
/// ```gemini
/// =>[<whitespace>]<URL>[<whitespace><USER-FRIENDLY LINK NAME>]
/// ```
///
/// where:
///
/// * `<whitespace>` is any non-zero number of consecutive spaces or tabs
/// * Square brackets indicate that the enclosed content is optional.
/// * `<URL>` is a URL, which may be absolute or relative. If the URL
///   does not include a scheme, a scheme of `gemini://` is implied.
class Link extends Node {
  final Uri to;
  final String? name;

  Link(this.to, this.name);

  @override
  String toString() => name == null ? "=> $to" : "=> $to $name";
}

/// Lines beginning with "* " are unordered list items. This line type exists
/// purely for stylistic reasons. The * may be replaced in advanced clients by
/// a bullet symbol. Any text after the "* " should be presented to the user as
/// if it were a text line, i.e. wrapped to fit the viewport and formatted
/// "nicely". Advanced clients can take the space of the bullet symbol into
/// account when wrapping long list items to ensure that all lines of text
/// corresponding to the item are offset an equal distance from the left of the screen.
class ListItem extends Node {
  final String body;

  ListItem(this.body);

  @override
  String toString() => "* $body";
}

/// Any line whose first three characters are "```" (i.e. three consecutive
/// back ticks with no leading whitespace) are preformatted toggle lines.
/// These lines should NOT be included in the rendered output shown to the
/// user. Instead, these lines toggle the parser between preformatted mode
/// being "on" or "off". Preformatted mode should be "off" at the beginning
/// of a document. The current status of preformatted mode is the only
/// internal state a parser is required to maintain. When preformatted mode
/// is "on", the usual rules for identifying line types are suspended, and
/// all lines should be identified as preformatted text lines (see 5.4.4).
///
/// Preformatted text lines should be presented to the user in a "neutral",
/// monowidth font without any alteration to whitespace or stylistic
/// enhancements. Graphical clients should use scrolling mechanisms to present
/// preformatted text lines which are longer than the client viewport, in
/// preference to wrapping. In displaying preformatted text lines, clients
/// should keep in mind applications like ASCII art and computer source
/// code: in particular, source code in languages with significant whitespace
/// (e.g. Python) should be able to be copied and pasted from the client into
/// a file and interpreted/compiled without any problems arising from the
/// client's manner of displaying them.
class Preformatted extends Node {
  final String? alt;
  final String text;

  Preformatted(this.alt, this.text);

  @override
  String toString() => "```${alt ?? ''}\n$text\n```";
}

/// Lines beginning with ">" are quote lines. This line type exists so that
/// advanced clients may use distinct styling to convey to readers the important
/// semantic information that certain text is being quoted from an external
/// source. For example, when wrapping long lines to the the viewport, each
/// resultant line may have a ">" symbol placed at the front.
class Quote extends Node {
  final String body;

  Quote(this.body);

  @override
  String toString() => "> $body";
}

/// Text lines are the most fundamental line type - any line which does not
/// match the definition of another line type defined below defaults to
/// being a text line. The majority of lines in a typical text/gemini document will be text lines.
class Text extends Node {
  final String text;

  Text(this.text);

  @override
  String toString() => text;
}
