import 'package:gemtext/gemtext.dart';
import 'package:test/test.dart';

const exampleText = """# Header
## Sub-header
### Sub-sub-header
=> gemini://example.com/  Link
* List item
> Quote
Plain text

```Alt
Preformatted
text
```""";

void main() {
  group('Parse', () {
    final doc = Gemtext.parse(exampleText);

    test('Header', () {
      Node node = doc.nodes[0];
      expect(node, isA<Heading>());

      node = node as Heading;
      expect(node.level, equals(HeadingLevel.h1));
      expect(node.body, equals("Header"));
    });

    test('Sub-header', () {
      Node node = doc.nodes[1];
      expect(node, isA<Heading>());

      node = node as Heading;
      expect(node.level, equals(HeadingLevel.h2));
      expect(node.body, equals("Sub-header"));
    });

    test('Sub-sub-header', () {
      Node node = doc.nodes[2];
      expect(node, isA<Heading>());

      node = node as Heading;
      expect(node.level, equals(HeadingLevel.h3));
      expect(node.body, equals("Sub-sub-header"));
    });

    test('Link', () {
      Node node = doc.nodes[3];
      expect(node, isA<Link>());

      node = node as Link;
      expect(node.to, equals(Uri.parse("gemini://example.com/")));
      expect(node.name, equals("Link"));
    });

    test('List item', () {
      Node node = doc.nodes[4];
      expect(node, isA<ListItem>());

      node = node as ListItem;
      expect(node.body, equals("List item"));
    });

    test('Quote', () {
      Node node = doc.nodes[5];
      expect(node, isA<Quote>());

      node = node as Quote;
      expect(node.body, equals("Quote"));
    });

    test('Plain text', () {
      Node node = doc.nodes[6];
      expect(node, isA<Text>());

      node = node as Text;
      expect(node.text, equals("Plain text"));
    });

    test('Empty plain text', () {
      Node node = doc.nodes[7];
      expect(node, isA<Text>());

      node = node as Text;
      expect(node.text, equals(""));
    });

    test('Preformatted', () {
      Node node = doc.nodes[8];
      expect(node, isA<Preformatted>());

      node = node as Preformatted;
      expect(node.alt, equals("Alt"));
      expect(node.text, equals("Preformatted\ntext"));
    });

    group('To string', () {
      void testToString(String name, Node node, String expected) {
        test(name, () {
          expect(node.toString(), equals(expected));
        });
      }

      testToString(
        "Header",
        Heading(HeadingLevel.h1, "Header"),
        "# Header",
      );
      testToString(
        "Sub-header",
        Heading(HeadingLevel.h2, "Sub-header"),
        "## Sub-header",
      );
      testToString(
        "Sub-sub-header",
        Heading(HeadingLevel.h3, "Sub-sub-header"),
        "### Sub-sub-header",
      );
      testToString(
        "Link",
        Link(Uri.parse("gemini://example.com/"), "Link"),
        "=> gemini://example.com/ Link",
      );
      testToString(
        "Link (no text)",
        Link(Uri.parse("gemini://example.com/"), null),
        "=> gemini://example.com/",
      );
      testToString(
        "List item",
        ListItem("List item"),
        "* List item",
      );
      testToString(
        "Quote",
        Quote("Quote"),
        "> Quote",
      );
      testToString(
        "Plain textt",
        Text("Plain text"),
        "Plain text",
      );
      testToString(
        "Preformatted (with alt)",
        Preformatted("Alt", "Preformatted\ntext"),
        "```Alt\nPreformatted\ntext\n```",
      );
      testToString(
        "Preformatted (without alt)",
        Preformatted(null, "Preformatted\ntext"),
        "```\nPreformatted\ntext\n```",
      );
    });
  });
}
